/*
 * Interface to PIC32 JTAG port using FT2232-based USB adapter.
 * For example: Olimex ARM-USB-Tiny adapter.
 *
 * Copyright (C) 2011 Serge Vakulenko
 *
 * This file is part of PIC32PROG project, which is distributed
 * under the terms of the GNU General Public License (GPL).
 * See the accompanying file "COPYING" for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#ifdef MINGW32
#include <windows.h>
#endif
#include "ftd2xx.h"
#ifndef FT_ERROR
#define FT_ERROR        (-1)
#endif

#include "adapter.h"
#include "pic32.h"

typedef struct {
    /* Common part */
    adapter_t adapter;
    const char *name;

    /* adapter mode */
    int mode; 
    
    /* Device handle for ftd2xx. */
    FT_HANDLE handle;

    /* adapter pin config and status  */
    unsigned short _dbus_tms;
    unsigned short _dbus_tck;
    unsigned short _dbus_tdi;
    unsigned short _dbus_tdo;
    unsigned short _dbus_sysrst;
    unsigned short _dbus_bufctl;
    unsigned short _dbus_ddr;
    unsigned short _dbus_data;

    /* Transmit buffer for MPSSE packet. */
    unsigned char output [256*64];
    int bytes_to_write;

    /* Receive buffer. */
    //unsigned char input [256*64];
    int bytes_to_drain;
    int bytes_to_drain_in;
    int bytes_to_read;
    int valid_bits;

    unsigned use_executable;
    unsigned serial_execution_mode;
} mpsse_adapter_t;

/*
 * Identifiers of USB adapter.
 */
#define FTDI_VID                0x0403
#define FTDI_PID_FT232R         0x6001
#define FTDI_PID_FT2232X        0x6010
#define FTDI_PID_FT232H         0x6014

#define OLIMEX_VID              0x15ba
#define OLIMEX_ARM_USB_TINY     0x0004  /* ARM-USB-Tiny */
#define OLIMEX_ARM_USB_TINY_H   0x002a	/* ARM-USB-Tiny-H */

/* MPSSE commands. */
#define CLKWNEG                 0x01
#define BITMODE                 0x02
#define CLKRNEG                 0x04
#define LSB                     0x08
#define WTDI                    0x10
#define RTDO                    0x20
#define WTMS                    0x40

#define SET_BITS_LOWBYTE	0x80
#define SET_BITS_HIGHBYTE	0x82
#define READ_BITS_LOWBYTE	0x81
#define READ_BITS_HIGHBYTE	0x83
#define DISABLE_LOOPBACK	0x85
#define ENABLE_LOOPBACK		0x84
#define SET_TCK_DIVISOR		0x86

    unsigned char enable_loopback[] = "\x85";

#define amode		(a->mode)
#define SYNCBB_JTAG	0
#define SYNCBB_ICSP	1
#define MPSSE_JTAG	2
#define MPSSE_ICSP	3
#define IS_MPSSE(m)	(!!((m)&2))
#define IS_SYNCBB(m)	(!((m)&2))
#define IS_ICSP(m)	(!!((m)&1))

#define dbus_tms	(a->_dbus_tms)
#define dbus_tck	(a->_dbus_tck)
#define dbus_tdi	(a->_dbus_tdi)
#define dbus_tdo	(a->_dbus_tdo)
#define dbus_sysrst	(a->_dbus_sysrst)
#define dbus_bufctl	(a->_dbus_bufctl)
#define dbus_ddr	(a->_dbus_ddr)
#define dbus_data	(a->_dbus_data)

static int h1toi(char c) {
	if (('0' <= c ) && ( c <= '9')) return c - '0';
	if (('a' <= c ) && ( c <= 'f')) return c - 'a';
	if (('A' <= c ) && ( c <= 'F')) return c - 'A';
	return 16;
}

static int setup_mode (mpsse_adapter_t *a, char *opt) {
	int err = 0;
	static char namebuf[256];
	if (!strcasecmp(opt, "mpsse-icsp")) {
		dbus_tck    = 1 << 0;	// ACBUS0
		dbus_tdi    = 1 << 1;	// ACBUS1
		dbus_tdo    = 0;
		dbus_tms    = 0;
		dbus_sysrst = 1 << 7;	// ACBUS7
		dbus_bufctl = 0;
		amode = MPSSE_ICSP;
	} else if (!strcasecmp(opt, "mpsse-jtag")) {
		dbus_tck    = 1 << 0;	// ACBUS0
		dbus_tdi    = 1 << 1;	// ACBUS1
		dbus_tdo    = 1 << 2;	// ACBUS2
		dbus_tms    = 1 << 3;	// ACBUS3
		dbus_sysrst = 1 << 7;	// ACBUS7
		dbus_bufctl = 0;
		amode = MPSSE_JTAG;
	} else if (!strcasecmp(opt, "syncbb-icsp") || !(*opt) ) {
		dbus_tck    = 1 << 5;	// D5/DSR
		dbus_tdi    = 1 << 6;	// D6/DCD
		dbus_tdo    = 1 << 3;	// D3/CTS
		dbus_tms    = 0;
		dbus_sysrst = 1 << 7;	// D7/RI
		dbus_bufctl = 0;
		amode = SYNCBB_ICSP;
	} else if (!strcasecmp(opt, "syncbb-jtag")) {
		dbus_tck    = 1 << 5;	// D5/DSR
		dbus_tdi    = 1 << 2;	// D2/RTS
		dbus_tdo    = 1 << 3;	// D3/CTS
		dbus_tms    = 1 << 7;	// D7/RI 
		dbus_sysrst = 1 << 4;	// D4/DTR
		dbus_bufctl = 0;
		amode = SYNCBB_JTAG;
	} else if (!strncasecmp(opt, "mpsse-icsp-",11) 
		&& ( (strlen(opt)==14) || (strlen(opt)==15))) {
		dbus_tck    = 1 << h1toi(opt[11]);
		dbus_tdi    = 1 << h1toi(opt[12]);
		dbus_tdo    = 0;
		dbus_tms    = 0;
		dbus_sysrst = 1 << h1toi(opt[13]);
		dbus_bufctl = 1 << h1toi(opt[14]);
		amode = MPSSE_ICSP;
	} else if (!strncasecmp(opt, "mpsse-jtag-",11)
		&& ( (strlen(opt)==16) || (strlen(opt)==17))) {
		dbus_tck    = 1 << h1toi(opt[11]);
		dbus_tdi    = 1 << h1toi(opt[12]);
		dbus_tdo    = 1 << h1toi(opt[13]);
		dbus_tms    = 1 << h1toi(opt[14]);
		dbus_sysrst = 1 << h1toi(opt[15]);
		dbus_bufctl = 1 << h1toi(opt[16]);
		amode = MPSSE_JTAG;
	} else if (!strncasecmp(opt, "syncbb-icsp-",12)
		&& ( (strlen(opt)==16) || (strlen(opt)==17))) {
		dbus_tck    = 1 << h1toi(opt[12]);
		dbus_tdi    = 1 << h1toi(opt[13]);
		dbus_tdo    = 1 << h1toi(opt[14]);
		dbus_tms    = 0;
		dbus_sysrst = 1 << h1toi(opt[15]);
		dbus_bufctl = 1 << h1toi(opt[16]);
		amode = SYNCBB_ICSP;
	} else if (!strncasecmp(opt, "syncbb-jtag-",12)
		&& ( (strlen(opt)==17) || (strlen(opt)==18))) {
		dbus_tck    = 1 << h1toi(opt[12]);
		dbus_tdi    = 1 << h1toi(opt[13]);
		dbus_tdo    = 1 << h1toi(opt[14]);
		dbus_tms    = 1 << h1toi(opt[15]);
		dbus_sysrst = 1 << h1toi(opt[16]);
		dbus_bufctl = 1 << h1toi(opt[17]);
		amode = SYNCBB_JTAG;
	} else {
		err = -1;
	}
	strcpy(namebuf, "noname");
	if (err == 0) {
		if (*opt) strcpy(namebuf, opt);
		else      strcpy(namebuf, "syncbb-icsp");
	}
	a->name = namebuf;
	dbus_ddr = dbus_tms | dbus_tck | dbus_tdi | dbus_sysrst | dbus_bufctl;
	dbus_data = dbus_sysrst;
	return err;
}

/*
 * Calculate checksum.
 */
static unsigned calculate_crc (unsigned crc, unsigned char *data, unsigned nbytes)
{
    static const unsigned short crc_table [16] = {
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
        0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    };
    unsigned i;

    while (nbytes--) {
        i = (crc >> 12) ^ (*data >> 4);
        crc = crc_table[i & 0x0F] ^ (crc << 4);
        i = (crc >> 12) ^ (*data >> 0);
        crc = crc_table[i & 0x0F] ^ (crc << 4);
        data++;
    }
    return crc & 0xffff;
}

static void ftdi_dump(char *title, unsigned char *buf, int len, int off, int incl, int n_bits) {
	int i;
	int idx = off; 
	if (n_bits < 16) {
		fprintf(stderr, "\t%s :", title);
		for (i=0; i<n_bits; i++) {
			fprintf(stderr, "%02x ", buf[idx]);
			idx += incl;
		}
		fprintf(stderr, "\n");
	}
}

/*
 * If there are any data in transmit buffer -
 * send them to device.
 */
static void mpsse_flush_output (mpsse_adapter_t *a)
{
    DWORD rlen;

    if (a->bytes_to_write <= 0) {
        if (a->bytes_to_drain >=0)
	    goto force_flush;
	return;
    }

    if (debug_level) {
    	fflush(stdout);
    	fflush(stderr);
    }
    FT_Write(a->handle, a->output, a->bytes_to_write, &rlen);
    if (debug_level>1) {
           fprintf (stderr, " write : to_write %d to_drain %d to_read %d v %d rlen %d\n", 
			a->bytes_to_write,
			a->bytes_to_drain + a->bytes_to_drain_in, 
			a->bytes_to_read, 
			a->valid_bits, (int)rlen);
    }
    a->bytes_to_write = 0;
    a->bytes_to_drain += a->bytes_to_drain_in;
    a->bytes_to_drain_in = 0;

    if (a->bytes_to_read <= 0) {
	if (a->bytes_to_drain >= sizeof(a->output)/2) {
force_flush:
		if (debug_level>1) {
			fprintf (stderr, " drain : to_drain %d ", a->bytes_to_drain);
		}
		if (debug_level) {
			fflush(stderr);
		}
    		FT_Read(a->handle, a->output, a->bytes_to_drain, &rlen);
                if (debug_level>1) {
           		fprintf (stderr, "rlen %d\n", (int)rlen);
                }
    		a->bytes_to_drain = 0;
	}
    }
}

static void ftdi_putc (mpsse_adapter_t *a, unsigned char v)
{
    a->output[a->bytes_to_write++] = v;
}

static unsigned long long mpsse_recv (mpsse_adapter_t *a) 
{
    DWORD rlen;
    int i, n_bits;
    unsigned long long rval = 0L;
    unsigned long long mask = 1L;

    mpsse_flush_output (a);
    mpsse_flush_output (a);
   if (a->bytes_to_read <= 0) return rval;

    n_bits = a->valid_bits;
    if (debug_level>2) {
           fprintf (stderr, " read : to_read %d v %d", a->bytes_to_read, n_bits);
    }
    if (debug_level) {
	fflush(stdout);
	fflush(stderr);
    }
    FT_Read(a->handle, a->output, a->bytes_to_read, &rlen);
    if (debug_level>2) {
           fprintf (stderr, " rlen %d\n", (int)rlen);
    }
    a->bytes_to_read = 0;
    a->valid_bits = 0;
    switch (amode) {
    case MPSSE_JTAG:
	if (debug_level>2) {
	   ftdi_dump("", a->output, (int)rlen, 0, 1, n_bits);
	}
	for (i=0; i< n_bits; i++) {
		if (a->output[i] & dbus_tdo) 
			rval |= mask;
		mask <<= 1;
	}
	break;
    case MPSSE_ICSP:
	if (debug_level>2) {
	   ftdi_dump("", a->output, (int)rlen, 0, 1, n_bits);
	}
	for (i=0; i< n_bits; i++) {
		if (a->output[i] & dbus_tdi) 
			rval |= mask;
		mask <<= 1;
	}
	break;
    case SYNCBB_JTAG:
	if (debug_level>2) {
	   ftdi_dump("", a->output, (int)rlen, 0, 2, n_bits);
	}
	for (i=0; i< n_bits; i++) {
		if (a->output[i*2] & dbus_tdo) 
			rval |= mask;
		mask <<= 1;
	}
	break;
    case SYNCBB_ICSP:
	if (debug_level>2) {
	   ftdi_dump("", a->output, (int)rlen, 6, 8, n_bits);
	}
	for (i=0; i< n_bits; i++) {
		if (a->output[i*8+6] & dbus_tdo)
			rval |= mask;
		mask <<= 1;
	}
    }
    return rval;
}

static void ftdi_send_1bit (mpsse_adapter_t *a, int tdi, int tms, int read_flag)
{
    if (debug_level>3) {
	if (tms) fprintf(stderr, "T");
	else if (tdi) fprintf(stderr, "1");
	else fprintf(stderr, "0");
    }
    if (read_flag & 0x2)
	a->valid_bits += 1;
    switch (amode) {
    case MPSSE_JTAG:
	if (read_flag) {
		a->bytes_to_read += 1;
		ftdi_putc(a, READ_BITS_HIGHBYTE);
	}
	if (tdi) dbus_data |= dbus_tdi;
	else     dbus_data &= ~dbus_tdi;
	if (tms) dbus_data |= dbus_tms;
	else     dbus_data &= ~dbus_tms;
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	break;
    case MPSSE_ICSP:
	if (tdi) dbus_data |= dbus_tdi;
	else     dbus_data &= ~dbus_tdi;
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	if (tms) dbus_data |= dbus_tdi;
	else     dbus_data &= ~dbus_tdi;
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);

	dbus_data |= dbus_tdi;
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_ddr & ~dbus_tdi & 0xff);
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & ~dbus_tdi & 0xff);
	if (read_flag) {
		a->bytes_to_read += 1;
		ftdi_putc(a, READ_BITS_HIGHBYTE);
	}
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_ddr & ~dbus_tdi & 0xff);
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & ~dbus_tdi & 0xff);
	break;
    case SYNCBB_JTAG:
        if (read_flag) a->bytes_to_read += 2;
	else           a->bytes_to_drain_in += 2;
	if (tdi) dbus_data |= dbus_tdi;
	else     dbus_data &= ~dbus_tdi;
	if (tms) dbus_data |= dbus_tms;
	else     dbus_data &= ~dbus_tms;
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_data & 0xff);
	break;
    case SYNCBB_ICSP:
        if (read_flag) a->bytes_to_read += 8;
	else           a->bytes_to_drain_in += 8;
	if (tdi) dbus_data &= ~dbus_tdi; // reversed 
	else     dbus_data |= dbus_tdi;
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_data & 0xff);
	if (tms) dbus_data &= ~dbus_tdi; // reversed 
	else     dbus_data |= dbus_tdi;
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_data & 0xff);
	dbus_data &= ~dbus_tdi;	// 0 : pulluped
	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_data & 0xff);

	ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
	ftdi_putc(a, dbus_data & 0xff);
	break;
    }
}

static void ftdi_send_key (mpsse_adapter_t *a, unsigned key)
{
    int i;
    if (debug_level>0) {
           fprintf (stderr, " ## ftdi_send_key 0x%x\n", key);
    }
    switch (amode) {
    case MPSSE_ICSP:
	for (i=0; i<32; i++) {
		if (key & 0x8000)  dbus_data |= dbus_tdi;
		else               dbus_data &= ~dbus_tdi;
		ftdi_putc(a, SET_BITS_HIGHBYTE);
		ftdi_putc(a, dbus_data & 0xff);
		ftdi_putc(a, dbus_ddr & 0xff);
		ftdi_putc(a, SET_BITS_HIGHBYTE);
		ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
		ftdi_putc(a, dbus_ddr & 0xff);
		key <<= 1;
	}
	dbus_data &= ~dbus_tdi;
	ftdi_putc(a, SET_BITS_HIGHBYTE);
	ftdi_putc(a, dbus_data & 0xff);
	ftdi_putc(a, dbus_ddr & 0xff);
	mpsse_flush_output (a);
	break;	
    case SYNCBB_ICSP:
	for (i=0; i<32; i++) {
		if (key & 0x8000)  dbus_data &= ~dbus_tdi; // reversed
		else               dbus_data |= dbus_tdi;
		a->bytes_to_drain_in += 2;
		ftdi_putc(a, dbus_data & 0xff);
		ftdi_putc(a, (dbus_data | dbus_tck) & 0xff);
		key <<= 1;
	}
	a->bytes_to_drain_in += 1;
	ftdi_putc(a, dbus_data & 0xff);
	mpsse_flush_output (a);
	break;
    }
}

static void mpsse_send (mpsse_adapter_t *a,
    unsigned tms_prolog_nbits, unsigned tms_prolog,
    unsigned tdi_nbits, unsigned long long tdi, int read_flag)
{
    unsigned tms_epilog_nbits = 0, tms_epilog = 0;
    int readflag_epi;
    int i;

    if (debug_level>1) {
           fprintf (stderr, " ## mpsse_send :" );
    }
    if (read_flag) read_flag = 0x3;

    if (tdi_nbits > 0) {
        /* We have some data; add generic prologue TMS 1-0-0
         * and epilogue TMS 1-0. */
        tms_prolog |= 1 << tms_prolog_nbits;
        tms_prolog_nbits += 3;
        tms_epilog = 1;
        tms_epilog_nbits = 2;
    }
    for (i=0; i< tms_prolog_nbits; i++) {
	ftdi_send_1bit (a, 0 , (tms_prolog & 0x1), 0);
	tms_prolog >>= 1;
    }
    for (i=0; i< tdi_nbits; i++) {
	ftdi_send_1bit(a, ((unsigned)tdi & 0x1), (i+1)==tdi_nbits , read_flag);
	tdi >>= 1;
    }
    readflag_epi = (IS_SYNCBB(amode) && read_flag)? 0x1:0;
    for (i=0; i< tms_epilog_nbits; i++) {
	ftdi_send_1bit (a, 0 , (tms_epilog & 0x1), readflag_epi);
	tms_epilog >>= 1;
    }
    if (debug_level>1) {
	   fprintf(stderr, "\n");
           fprintf (stderr, "\tto_write %d to_drain %d to_read %d v %d\n", 
           		a->bytes_to_write,
           		a->bytes_to_drain + a->bytes_to_drain_in,
           		a->bytes_to_read,
           		a->valid_bits);
    }
    if ( (a->bytes_to_write >= sizeof(a->output) - 24)
	  || (a->bytes_to_drain >= sizeof(a->output)/2) ) {
    	mpsse_flush_output (a);
    }
}

static void ftdi_reset (mpsse_adapter_t *a, int trst, int sysrst, int led)
{
    mpsse_flush_output (a);

    if (debug_level>0) {
           fprintf (stderr, " ## ftdi_reset %d\n", sysrst);
    }
    if (!sysrst) dbus_data |= dbus_sysrst;
    else         dbus_data &= ~dbus_sysrst;

    if (IS_MPSSE(amode)) {
		ftdi_putc(a, SET_BITS_LOWBYTE); // swapped
		ftdi_putc(a, (dbus_data >> 8)& 0xff);
		ftdi_putc(a, (dbus_ddr >> 8) & 0xff);

		ftdi_putc(a, SET_BITS_HIGHBYTE);
		ftdi_putc(a, dbus_data & 0xff);
		ftdi_putc(a, dbus_ddr & 0xff);
    } else { // IS_SYNCBB
		a->bytes_to_drain_in += 1;
		ftdi_putc(a, dbus_data & 0xff);
    }
    mpsse_flush_output (a);
}

static void mpsse_speed (mpsse_adapter_t *a, int divisor)
{
    /* command "set TCK divisor" */
    ftdi_putc(a, SET_TCK_DIVISOR);
    ftdi_putc(a, divisor & 0xff);
    ftdi_putc(a, (divisor >> 8) & 0xff);
    mpsse_flush_output (a);

    /* Disable TDI to TDO loopback. */
    ftdi_putc(a, DISABLE_LOOPBACK);
    mpsse_flush_output (a);
}

static void ftdi_speed (mpsse_adapter_t *a, int divisor)
{
    if (IS_MPSSE(amode)) {
	mpsse_speed(a, divisor);
    }
}

static void ftdi_close (adapter_t *adapter, int power_on)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;

    /* Clear EJTAGBOOT mode. */
    mpsse_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
    mpsse_send (a, 6, 31, 0, 0, 0);             /* TMS 1-1-1-1-1-0 */
    mpsse_flush_output (a);
    mpsse_flush_output (a); // force (for SYNCBB)

    /* Toggle /SYSRST. */
    ftdi_reset (a, 0, 1, 1);
    if (IS_MPSSE(amode)) {
    	dbus_data = 0;
    	dbus_ddr = 0;
    	ftdi_reset (a, 0, 0, 0);
    } else {
    	FT_SetBitMode(a->handle, 0, FT_BITMODE_SYNC_BITBANG);
    }
    FT_SetBitMode(a->handle, 0, FT_BITMODE_RESET);

    FT_SetLatencyTimer(a->handle, 16); // default
    FT_Close(a->handle);

    free (a);
}

/*
 * Read the Device Identification code
 */
static unsigned mpsse_get_idcode (adapter_t *adapter)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    unsigned idcode;

    /* Reset the JTAG TAP controller: TMS 1-1-1-1-1-0.
     * After reset, the IDCODE register is always selected.
     * Read out 32 bits of data. */
    mpsse_send (a, 6, 31, 32, 0, 1);
    idcode = mpsse_recv (a);
    return idcode;
}

/*
 * Put device to serial execution mode.
 */
static void serial_execution (mpsse_adapter_t *a)
{
    if (a->serial_execution_mode)
        return;
    a->serial_execution_mode = 1;

    /* Enter serial execution. */
    if (debug_level > 0)
        fprintf (stderr, "%s: enter serial execution\n", a->name);

    mpsse_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
    mpsse_send (a, 1, 1, 5, ETAP_EJTAGBOOT, 0); /* Send command. */

    /* Check status. */
    mpsse_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    mpsse_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    mpsse_send (a, 0, 0, 8, MCHP_DEASSERT_RST, 0);  /* Xfer data. */
    mpsse_send (a, 0, 0, 8, MCHP_FLASH_ENABLE, 0);  /* Xfer data. */
    mpsse_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
    unsigned status = mpsse_recv (a);
    if (debug_level > 0)
        fprintf (stderr, "%s: status %04x\n", a->name, status);
    if (status != (MCHP_STATUS_CPS | MCHP_STATUS_CFGRDY |
                   MCHP_STATUS_FAEN | MCHP_STATUS_DEVRST)) {
        fprintf (stderr, "%s: invalid status = %04x (reset)\n", a->name, status);
        exit (-1);
    }

    /* Deactivate /SYSRST. */
    ftdi_reset (a, 0, 0, 1);
    mdelay (10);

    /* Check status. */
    mpsse_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    mpsse_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    mpsse_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
    status = mpsse_recv (a);
    if (debug_level > 0)
        fprintf (stderr, "%s: status %04x\n", a->name, status);
    if (status != (MCHP_STATUS_CPS | MCHP_STATUS_CFGRDY |
                   MCHP_STATUS_FAEN)) {
        fprintf (stderr, "%s: invalid status = %04x (no reset)\n", a->name, status);
        exit (-1);
    }

    /* Leave it in ETAP mode. */
    mpsse_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
    mpsse_flush_output (a);
}

static void xfer_fastdata (mpsse_adapter_t *a, unsigned word)
{
    mpsse_send (a, 0, 0, 33, (unsigned long long) word << 1, 0);
}

static void xfer_instruction (mpsse_adapter_t *a, unsigned instruction)
{
    unsigned ctl;

    if (debug_level > 1)
        fprintf (stderr, "%s: xfer instruction %08x\n", a->name, instruction);

    // Select Control Register
    mpsse_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */

    // Wait until CPU is ready
    // Check if Processor Access bit (bit 18) is set
    do {
        mpsse_send (a, 0, 0, 32, CONTROL_PRACC |    /* Xfer data. */
                                 CONTROL_PROBEN |
                                 CONTROL_PROBTRAP |
                                 CONTROL_EJTAGBRK, 1);
        ctl = mpsse_recv (a);
    } while (! (ctl & CONTROL_PRACC));

    // Select Data Register
    // Send the instruction
    mpsse_send (a, 1, 1, 5, ETAP_DATA, 0);          /* Send command. */
    mpsse_send (a, 0, 0, 32, instruction, 0);       /* Send data. */

    // Tell CPU to execute instruction
    mpsse_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */
    mpsse_send (a, 0, 0, 32, CONTROL_PROBEN |       /* Send data. */
                             CONTROL_PROBTRAP, 0);
}

static unsigned get_pe_response (mpsse_adapter_t *a)
{
    unsigned ctl, response;

    // Select Control Register
    mpsse_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */

    // Wait until CPU is ready
    // Check if Processor Access bit (bit 18) is set
    do {
        mpsse_send (a, 0, 0, 32, CONTROL_PRACC |    /* Xfer data. */
                                 CONTROL_PROBEN |
                                 CONTROL_PROBTRAP |
                                 CONTROL_EJTAGBRK, 1);
        ctl = mpsse_recv (a);
    } while (! (ctl & CONTROL_PRACC));

    // Select Data Register
    // Send the instruction
    mpsse_send (a, 1, 1, 5, ETAP_DATA, 0);          /* Send command. */
    mpsse_send (a, 0, 0, 32, 0, 1);                 /* Get data. */
    response = mpsse_recv (a);

    // Tell CPU to execute NOP instruction
    mpsse_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */
    mpsse_send (a, 0, 0, 32, CONTROL_PROBEN |       /* Send data. */
                             CONTROL_PROBTRAP, 0);
    if (debug_level > 1)
        fprintf (stderr, "%s: get PE response %08x\n", a->name, response);
    return response;
}

/*
 * Read a word from memory (without PE).
 */
static unsigned mpsse_read_word (adapter_t *adapter, unsigned addr)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    unsigned addr_lo = addr & 0xFFFF;
    unsigned addr_hi = (addr >> 16) & 0xFFFF;

    serial_execution (a);

    //fprintf (stderr, "%s: read word from %08x\n", a->name, addr);
    xfer_instruction (a, 0x3c04bf80);           // lui s3, 0xFF20
    xfer_instruction (a, 0x3c080000 | addr_hi); // lui t0, addr_hi
    xfer_instruction (a, 0x35080000 | addr_lo); // ori t0, addr_lo
    xfer_instruction (a, 0x8d090000);           // lw t1, 0(t0)
    xfer_instruction (a, 0xae690000);           // sw t1, 0(s3)

    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    mpsse_send (a, 0, 0, 33, 0, 1);             /* Get fastdata. */
    unsigned word = mpsse_recv (a) >> 1;

    if (debug_level > 0)
        fprintf (stderr, "%s: read word at %08x -> %08x\n", a->name, addr, word);
    return word;
}

/*
 * Read a memory block.
 */
static void mpsse_read_data (adapter_t *adapter,
    unsigned addr, unsigned nwords, unsigned *data)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    unsigned words_read, i;

    //fprintf (stderr, "%s: read %d bytes from %08x\n", a->name, nwords*4, addr);
    if (! a->use_executable) {
        /* Without PE. */
        for (; nwords > 0; nwords--) {
            *data++ = mpsse_read_word (adapter, addr);
            addr += 4;
        }
        return;
    }

    /* Use PE to read memory. */
    for (words_read = 0; words_read < nwords; words_read += 32) {

        mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);
        xfer_fastdata (a, PE_READ << 16 | 32);      /* Read 32 words */
        xfer_fastdata (a, addr);                    /* Address */

        unsigned response = get_pe_response (a);    /* Get response */
        if (response != PE_READ << 16) {
            fprintf (stderr, "%s: bad READ response = %08x, expected %08x\n",
                a->name, response, PE_READ << 16);
            exit (-1);
        }
        for (i=0; i<32; i++) {
            *data++ = get_pe_response (a);          /* Get data */
        }
        addr += 32*4;
    }
}

/*
 * Download programming executable (PE).
 */
static void mpsse_load_executable (adapter_t *adapter,
    const unsigned *pe, unsigned nwords, unsigned pe_version)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;

    a->use_executable = 1;
    serial_execution (a);

    if (debug_level > 0)
        fprintf (stderr, "%s: download PE loader\n", a->name);

    /* Step 1. */
    xfer_instruction (a, 0x3c04bf88);   // lui a0, 0xbf88
    xfer_instruction (a, 0x34842000);   // ori a0, 0x2000 - address of BMXCON
    xfer_instruction (a, 0x3c05001f);   // lui a1, 0x1f
    xfer_instruction (a, 0x34a50040);   // ori a1, 0x40   - a1 has 001f0040
    xfer_instruction (a, 0xac850000);   // sw  a1, 0(a0)  - BMXCON initialized

    /* Step 2. */
    xfer_instruction (a, 0x34050800);   // li  a1, 0x800  - a1 has 00000800
    xfer_instruction (a, 0xac850010);   // sw  a1, 16(a0) - BMXDKPBA initialized

    /* Step 3. */
    xfer_instruction (a, 0x8c850040);   // lw  a1, 64(a0) - load BMXDMSZ
    xfer_instruction (a, 0xac850020);   // sw  a1, 32(a0) - BMXDUDBA initialized
    xfer_instruction (a, 0xac850030);   // sw  a1, 48(a0) - BMXDUPBA initialized

    /* Step 4. */
    xfer_instruction (a, 0x3c04a000);   // lui a0, 0xa000
    xfer_instruction (a, 0x34840800);   // ori a0, 0x800  - a0 has a0000800

    /* Download the PE loader. */
    int i;
    for (i=0; i<PIC32_PE_LOADER_LEN; i+=2) {
        /* Step 5. */
        unsigned opcode1 = 0x3c060000 | pic32_pe_loader[i];
        unsigned opcode2 = 0x34c60000 | pic32_pe_loader[i+1];

        xfer_instruction (a, opcode1);      // lui a2, PE_loader_hi++
        xfer_instruction (a, opcode2);      // ori a2, PE_loader_lo++
        xfer_instruction (a, 0xac860000);   // sw  a2, 0(a0)
        xfer_instruction (a, 0x24840004);   // addiu a0, 4
    }

    /* Jump to PE loader (step 6). */
    xfer_instruction (a, 0x3c19a000);   // lui t9, 0xa000
    xfer_instruction (a, 0x37390800);   // ori t9, 0x800  - t9 has a0000800
    xfer_instruction (a, 0x03200008);   // jr  t9
    xfer_instruction (a, 0x00000000);   // nop

    /* Switch from serial to fast execution mode. */
    mpsse_send (a, 1, 1, 5, TAP_SW_ETAP, 0);
    mpsse_send (a, 6, 31, 0, 0, 0);             /* TMS 1-1-1-1-1-0 */

    /* Send parameters for the loader (step 7-A).
     * PE_ADDRESS = 0xA000_0900,
     * PE_SIZE */
    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, 0xa0000900);
    xfer_fastdata (a, nwords);

    /* Download the PE itself (step 7-B). */
    if (debug_level > 0)
        fprintf (stderr, "%s: download PE\n", a->name);
    for (i=0; i<nwords; i++) {
        xfer_fastdata (a, *pe++);
    }
    mpsse_flush_output (a);
    mdelay (10);

    /* Download the PE instructions. */
    xfer_fastdata (a, 0);                       /* Step 8 - jump to PE. */
    xfer_fastdata (a, 0xDEAD0000);
    mpsse_flush_output (a);
    mdelay (10);
    xfer_fastdata (a, PE_EXEC_VERSION << 16);

    unsigned version = get_pe_response (a);
    if (version != (PE_EXEC_VERSION << 16 | pe_version)) {
        fprintf (stderr, "%s: bad PE version = %08x, expected %08x\n",
            a->name, version, PE_EXEC_VERSION << 16 | pe_version);
        exit (-1);
    }
    if (debug_level > 0)
        fprintf (stderr, "%s: PE version = %04x\n",
            a->name, version & 0xffff);
}

/*
 * Erase all flash memory.
 */
static void mpsse_erase_chip (adapter_t *adapter)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;

    mpsse_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    mpsse_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    mpsse_send (a, 0, 0, 8, MCHP_ERASE, 0);     /* Xfer data. */
    mpsse_flush_output (a);
    mdelay (400);

    /* Leave it in ETAP mode. */
    mpsse_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
}

/*
 * Write a word to flash memory.
 */
static void mpsse_program_word (adapter_t *adapter,
    unsigned addr, unsigned word)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;

    if (debug_level > 0)
        fprintf (stderr, "%s: program word at %08x: %08x\n", a->name, addr, word);
    if (! a->use_executable) {
        /* Without PE. */
        fprintf (stderr, "%s: slow flash write not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to write flash memory. */
    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_WORD_PROGRAM << 16 | 2);
    mpsse_flush_output (a);
    xfer_fastdata (a, addr);                    /* Send address. */
    mpsse_flush_output (a);
    xfer_fastdata (a, word);                    /* Send word. */

    unsigned response = get_pe_response (a);
    if (response != (PE_WORD_PROGRAM << 16)) {
        fprintf (stderr, "%s: failed to program word %08x at %08x, reply = %08x\n",
            a->name, word, addr, response);
        exit (-1);
    }
}

/*
 * Flash write row of memory.
 */
static void mpsse_program_row32 (adapter_t *adapter, unsigned addr,
    unsigned *data)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    int i;

    if (debug_level > 0)
        fprintf (stderr, "%s: row program 128 bytes at %08x\n", a->name, addr);
    if (! a->use_executable) {
        /* Without PE. */
        fprintf (stderr, "%s: slow flash write not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to write flash memory. */
    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_ROW_PROGRAM << 16 | 32);
    mpsse_flush_output (a);
    xfer_fastdata (a, addr);                    /* Send address. */

    /* Download 128 bytes of data. */
    for (i = 0; i < 32; i++) {
        if ((i & 7) == 0)
            mpsse_flush_output (a);
        xfer_fastdata (a, *data++);             /* Send word. */
    }
    mpsse_flush_output (a);

    unsigned response = get_pe_response (a);
    if (response != (PE_ROW_PROGRAM << 16)) {
        fprintf (stderr, "%s: failed to program row at %08x, reply = %08x\n",
            a->name, addr, response);
        exit (-1);
    }
}

static void mpsse_program_row128 (adapter_t *adapter, unsigned addr,
    unsigned *data)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    int i;

    if (debug_level > 0)
        fprintf (stderr, "%s: row program 512 bytes at %08x\n", a->name, addr);
    if (! a->use_executable) {
        /* Without PE. */
        fprintf (stderr, "%s: slow flash write not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to write flash memory. */
    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_ROW_PROGRAM << 16 | 128);
    mpsse_flush_output (a);
    xfer_fastdata (a, addr);                    /* Send address. */

    /* Download 512 bytes of data. */
    for (i = 0; i < 128; i++) {
        if ((i & 7) == 0)
            mpsse_flush_output (a);
        xfer_fastdata (a, *data++);             /* Send word. */
    }
    mpsse_flush_output (a);

    unsigned response = get_pe_response (a);
    if (response != (PE_ROW_PROGRAM << 16)) {
        fprintf (stderr, "%s: failed to program row at %08x, reply = %08x\n",
            a->name, addr, response);
        exit (-1);
    }
}

/*
 * Verify a block of memory.
 */
static void mpsse_verify_data (adapter_t *adapter,
    unsigned addr, unsigned nwords, unsigned *data)
{
    mpsse_adapter_t *a = (mpsse_adapter_t*) adapter;
    unsigned data_crc, flash_crc;

    //fprintf (stderr, "%s: verify %d words at %08x\n", a->name, nwords, addr);
    if (! a->use_executable) {
        /* Without PE. */
        fprintf (stderr, "%s: slow verify not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to get CRC of flash memory. */
    mpsse_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_GET_CRC << 16);
    mpsse_flush_output (a);
    xfer_fastdata (a, addr);                    /* Send address. */
    mpsse_flush_output (a);
    xfer_fastdata (a, nwords * 4);              /* Send length. */

    unsigned response = get_pe_response (a);
    if (response != (PE_GET_CRC << 16)) {
        fprintf (stderr, "%s: failed to verify %d words at %08x, reply = %08x\n",
            a->name, nwords, addr, response);
        exit (-1);
    }
    flash_crc = get_pe_response (a) & 0xffff;

    data_crc = calculate_crc (0xffff, (unsigned char*) data, nwords * 4);
    if (flash_crc != data_crc) {
        fprintf (stderr, "%s: checksum failed at %08x: sum=%04x, expected=%04x\n",
            a->name, addr, flash_crc, data_crc);
        //exit (-1);
    }
}

/*
 * Initialize adapter F2232.
 * Return a pointer to a data structure, allocated dynamically.
 * When adapter not found, return 0.
 */
adapter_t *adapter_open_ftdi (char *opt)
{
    mpsse_adapter_t *a;
    char *p, port[256];
    char name[256];
    int r;
    FT_DEVICE ft_type;
    DWORD ft_id;
    DWORD trans_size = 16384 * 4;
    char sn[16];
    char desc[64];

    a = calloc (1, sizeof (*a));
    if (! a) {
        fprintf (stderr, "%s: out of memory\n", name);
        return 0;
    }
    port[0] = 0;
    name[0] = 0;
    if (opt && *opt) {
        strcpy(port, opt);
        p = strchr(port, '-');
        if (p) {
	    *p = 0;
	    strcpy(name, p+1);
	}
	if (!strcasecmp(port, "mpsse") || !strcasecmp(port, "syncbb")) {
		strcpy(name, opt);
		port[0] = 0;
	}
    }
    r = FT_ERROR;
#if 0
    DWORD pid,vid;
    if (sscanf(port,"%04x:%04x",&vid, &pid) == 2) {
	FT_SetVIDPID(vid, pid);
	port[0] = 0;
    }
#endif
    if (!strncasecmp("ft", port, 2) &&  '0' <= port[2] && port[2] <= '9') {
	r = FT_Open(port[2] - '0', &(a->handle));
    }
    if ((r != FT_OK) && port[0]) {
	r = FT_OpenEx(port, FT_OPEN_BY_DESCRIPTION, &(a->handle));
	if (r != FT_OK)
	    r = FT_OpenEx(port, FT_OPEN_BY_SERIAL_NUMBER, &(a->handle));
    }
    if ((r != FT_OK) && !port[0]) {
	r = FT_Open(0, &(a->handle));
    }
    if (r != FT_OK) {
	if (debug_level)
	    fprintf (stderr, "FTDI adapter not found: \n");
        free (a);
        return 0;
    }
    if (setup_mode(a, name)) {
        free (a);
	return 0;
    }

    FT_SetBitMode(a->handle, 0, FT_BITMODE_RESET);
    r = FT_GetDeviceInfo(a->handle, &ft_type, &ft_id, sn, desc, NULL);
    if (r != FT_OK) {
        fprintf (stderr, "FTDI adapter: can't get DeviceInfo \n");
        free (a);
        return 0;
    }
    if (ft_type <= FT_DEVICE_UNKNOWN) {
        fprintf (stderr, "FTDI adapter: (%s) is too old device\n", desc);
        free (a);
	return 0;
    }
    if ( (ft_type == FT_DEVICE_232R) && IS_MPSSE(amode)) {
        fprintf (stderr, "FTDI adapter: (%s) doesn't support MPSSE\n", desc);
        free (a);
	return 0;
    }
    if (debug_level) {
        fprintf (stderr, "FTDI adapter: found (%s) mode (%s)\n", desc,a->name);
    }
    if (ft_type >= FT_DEVICE_2232H) {
	FT_SetUSBParameters(a->handle, trans_size, trans_size);
	FT_SetBaudRate(a->handle, 1000000);
    } else {
	FT_SetBaudRate(a->handle, 115200);
    }

    if (IS_MPSSE(amode)) {
	r = FT_SetBitMode(a->handle, 0, FT_BITMODE_MPSSE);
        if ( (r != FT_OK) && debug_level) {
           fprintf (stderr, "FTDI adapter: set MPSSE mode failed\n");
        }
    } else {
	r = FT_SetBitMode(a->handle, dbus_ddr & 0xff, FT_BITMODE_SYNC_BITBANG);
        if ( (r != FT_OK) && debug_level) {
           fprintf (stderr, "FTDI adapter: set SYNCBB mode failed\n");
        }

    }
    /* Optimal rate is 0.5 MHz.
     * Divide base oscillator 6 MHz by 12. */
    unsigned divisor = 12 - 1;
    unsigned char latency_timer = 1;

     FT_SetLatencyTimer(a->handle, latency_timer);
    if (debug_level) {
    	fprintf (stderr, "%s: divisor: %u\n", a->name, divisor);
    	fprintf (stderr, "%s: latency timer: %u usec\n", a->name, latency_timer);
    }
    if (debug_level) {
     int baud = 6000000 / (divisor + 1);
        fprintf (stderr, "%s: speed %d samples/sec\n", a->name, baud);
    }
    ftdi_speed (a, divisor);

    if (IS_ICSP(amode)) {
    	ftdi_reset (a, 0, 1, 1);
        ftdi_reset (a, 0, 0, 1);
    	ftdi_reset (a, 0, 1, 1);
	ftdi_send_key(a, 0x4D434850);
    	ftdi_reset (a, 0, 0, 1);
    } else {
        ftdi_reset (a, 0, 0, 1);
        /* Activate /SYSRST. (L level)*/
        ftdi_reset (a, 0, 1, 1);
    }
    mpsse_flush_output (a);
    mpsse_flush_output (a);

    mdelay (10);

    if (debug_level > 1) debug_level +=2;
    /* Reset the JTAG TAP controller. */
    mpsse_send (a, 6, 31, 0, 0, 0);             /* TMS 1-1-1-1-1-0 */

    /* Check status. */
    mpsse_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    mpsse_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    mpsse_send (a, 0, 0, 8, MCHP_FLASH_ENABLE, 0);  /* Xfer data. */
    mpsse_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
    unsigned status = mpsse_recv (a);
    if (debug_level > 1 +2) debug_level -=2;
    if (debug_level > 0)
        fprintf (stderr, "%s: status %04x\n", a->name, status);
/*
 * MCHP status value.

 MCHP_STATUS_CPS    0x80 // Device is NOT code-protected
 MCHP_STATUS_NVMERR 0x20 // Error occured during NVM operation
 MCHP_STATUS_CFGRDY 0x08 // Configuration has been read and
                                // Code-Protect State bit is valid
 MCHP_STATUS_FCBUSY 0x04 // Flash Controller is Busy (erase is in progres
s)
 MCHP_STATUS_FAEN   0x02 // Flash access is enabled
 MCHP_STATUS_DEVRST 0x01 // Device reset is active
*/

    if (status != (MCHP_STATUS_CPS | MCHP_STATUS_CFGRDY |
                   MCHP_STATUS_FAEN | MCHP_STATUS_DEVRST)) { // 8B 
        fprintf (stderr, "%s: invalid status = %04x\n", a->name, status);
        free (a);
        return 0;
    }
    printf ("      Adapter: %s\n", a->name);

    /* User functions. */
    a->adapter.close = ftdi_close;
    a->adapter.get_idcode = mpsse_get_idcode;
    a->adapter.load_executable = mpsse_load_executable;
    a->adapter.read_word = mpsse_read_word;
    a->adapter.read_data = mpsse_read_data;
    a->adapter.verify_data = mpsse_verify_data;
    a->adapter.erase_chip = mpsse_erase_chip;
    a->adapter.program_word = mpsse_program_word;
    a->adapter.program_row32 = mpsse_program_row32;
    a->adapter.program_row128 = mpsse_program_row128;
    return &a->adapter;
}
